// lcd.c

// LCD data pins
#define LCD_DB0   PIN_B0
#define LCD_DB1   PIN_B1
#define LCD_DB2   PIN_B2
#define LCD_DB3   PIN_B3
#define LCD_DB4   PIN_B4
#define LCD_DB5   PIN_B5
#define LCD_DB6   PIN_B6
#define LCD_DB7   PIN_B7
// LCD control pins
#define LCD_E     PIN_E1
#define LCD_RS    PIN_E0
#define LCD_RW    PIN_E2

//========================================

#define lcd_type 2        // 0=5x7, 1=5x10, 2=2 lines
#define lcd_line_two 0x40 // LCD RAM address for the 2nd line


int8 const LCD_INIT_STRING[4] =
{
 0x30 | (lcd_type << 2), // Func set: 8-bit, 2 lines, 5x8 dots
 0x0C,                   // Display on
 0x01,                   // Clear display
 0x06                    // Increment cursor
 };
                             
//-------------------------------------
void lcd_send_nibble(int8 nibble)
{
 output_bit(LCD_DB0, !!(nibble & 1));
 output_bit(LCD_DB1, !!(nibble & 2)); 
 output_bit(LCD_DB2, !!(nibble & 4));   
 output_bit(LCD_DB3, !!(nibble & 8));   
 output_bit(LCD_DB4, !!(nibble & 16));
 output_bit(LCD_DB5, !!(nibble & 32)); 
 output_bit(LCD_DB6, !!(nibble & 64));   
 output_bit(LCD_DB7, !!(nibble & 128));   
 // E pulse
 delay_cycles(1);
 output_high(LCD_E);
 delay_us(2);
 output_low(LCD_E);
}

//-----------------------------------
// This sub-routine is only called by lcd_read_byte().
// It's not a stand-alone routine.  For example, the
// R/W signal is set high by lcd_read_byte() before
// this routine is called.     

int8 lcd_read_nibble(void)
{
  int8 retval;
  // Create bit variables so that we can easily set individual bits in the retval variable.
  #bit retval_0 = retval.0
  #bit retval_1 = retval.1
  #bit retval_2 = retval.2
  #bit retval_3 = retval.3
  #bit retval_4 = retval.4
  #bit retval_5 = retval.5
  #bit retval_6 = retval.6
  #bit retval_7 = retval.7
  retval = 0;
  // set E high  
  output_high(LCD_E);
  delay_cycles(1);
  // Read the value
  retval_0 = input(LCD_DB0);
  retval_1 = input(LCD_DB1);
  retval_2 = input(LCD_DB2);
  retval_3 = input(LCD_DB3);
  retval_4 = input(LCD_DB4);
  retval_5 = input(LCD_DB5);
  retval_6 = input(LCD_DB6);
  retval_7 = input(LCD_DB7);
  // set E low
  output_low(LCD_E);
  // Return the read value
  return(retval);   
}   

//---------------------------------------
// Read a byte from the LCD and return it.
int8 lcd_read_byte(void)
{
  output_high(LCD_RW);
  delay_cycles(1);
  return(lcd_read_nibble());
}

//----------------------------------------
// Send a byte to the LCD.
void lcd_send_byte(int8 address, int8 n)
{
  output_low(LCD_RS);
  // Wait until busy flag is low.
  while(bit_test(lcd_read_byte(),7)) ;

  if(address)
    output_high(LCD_RS);
  else
    output_low(LCD_RS);
  delay_cycles(1);
  output_low(LCD_RW);
  delay_cycles(1);
  output_low(LCD_E);
  lcd_send_nibble(n);
}

//----------------------------
void lcd_init(void)
{
  int8 i;
  output_low(LCD_RS);
  output_low(LCD_RW);
  output_low(LCD_E);
  delay_ms(15);

  for(i=0; i < 3; i++)
  {
    lcd_send_nibble(0x03);
    delay_ms(5);
  }
  lcd_send_nibble(0x02);
  // Init
  for(i=0; i < sizeof(LCD_INIT_STRING); i++)
  {
    lcd_send_byte(0, LCD_INIT_STRING[i]);
  }
}

//----------------------------
void lcd_gotoxy(int8 x, int8 y)
{
  int8 address;

  if(y != 1)
    address = lcd_line_two;
  else
    address=0;

  address += x-1;
  lcd_send_byte(0, 0x80 | address);
}

//-----------------------------
void lcd_putc(char c)
{
  switch(c)
  {
    case '\f':          // Clear display
      lcd_send_byte(0,1);
      delay_ms(2);
      break;
    case '\n':			// New line
      lcd_gotoxy(1,2);
      break;
    case '\b':          // Move back one position
      lcd_send_byte(0,0x10);
      break;
    default:
      lcd_send_byte(1,c);
      break;
  }
}

void lcd_writestring(char *s)
{
  while(*s)
  {
    lcd_putc(*s++);
  }
}

//------------------------------
char lcd_getc(int8 x, int8 y)
{
  char value;

  lcd_gotoxy(x,y);
  // Wait until busy flag is low.
  while(bit_test(lcd_read_byte(),7)); 
  output_high(LCD_RS);
  value = lcd_read_byte();
  output_low(lcd_RS);
  return(value);
}
